/**
 * @file main.c
 * @brief Este projeto demonstra o uso do modulo UARTx para ecoar caracteres digitados 
 * num Terminal via comunicacao serial assincrona.
 * @author Wu Shin-Ting
 * @date 23/11/2020
 *
 */

#include "derivative.h"

//==================================
// Clocks
//==================================
/**
 * @brief Setar bus clock em 20,97152MHz/[OUTDIV4]
 */
void SIM_initBusClock (uint8_t OUTDIV4) {	
	SIM_CLKDIV1 &= ~SIM_CLKDIV1_OUTDIV4(0b111);  //resetar todos os bits
	SIM_CLKDIV1 |= SIM_CLKDIV1_OUTDIV4(OUTDIV4); //mascarar 1s
}

/**
 * @brief Selecionar a base de tempo para os clocks dos modulos do microcontrolador
 * @param[in] base (MCGFLLCLK=0; MCGPLLCLK=1)
 */
void SIM_selPLLFLL (uint8_t base) {
	SIM_SOPT2 &= ~SIM_SOPT2_PLLFLLSEL_MASK;
	SIM_SOPT2 |= (base << SIM_SOPT2_PLLFLLSEL_SHIFT) & SIM_SOPT2_PLLFLLSEL_MASK;
}

//==================================
// PORTx
//==================================
void PORTx_muxUART0RxTxPins(void) {
	SIM_SCGC5 |= SIM_SCGC5_PORTA_MASK;      // habilitar clock gate de PORTA
	PORTA_PCR1 |= PORT_PCR_MUX(0x2);        // configurar com funcao UART0_RX
	PORTA_PCR2 |= PORT_PCR_MUX(0x2);        // configurar com funcao UART0_TX
}	

//==================================
// UART0
//==================================
/**
 * @Parametros de configuracao  UART0
 */
typedef struct _UART0Configuration {
	uint8_t	bdh_sbns;		// selecionar quantidade de stop bits (0 = 1; 1 = 2)
	uint16_t sbr;		
	uint8_t c1_loops;       // operacao em loop (normal = 0)
	uint8_t c1_dozeen;		// habilitar espera (doze)
	uint8_t c1_rsrc;		// habilitar a saida do TX em operao de loop
	uint8_t c1_m; 			// 8-bit (0) ou 9-bit de dados
	uint8_t c1_wake;		// forma de wakeup do RX (0-idle line; 1-address-mark)
	uint8_t c1_ilt;			// selecionar a forma de deteccao de "idle line"
	uint8_t c1_pe; 			// habilitar paridade
	uint8_t c1_pt;			// tipo de paridade (0-par; 1-impar)
	uint8_t c2_rwu; 		// setar o RX no estado de standby aguardando pelo wakeup
	uint8_t c2_sbk;			// habilitar o enfileiramento de caracteres break
	uint8_t	s2_msbf;		// setar endianess para MSB (0 = LSB;1 = MSB)
	uint8_t s2_rxinv;		// habilitar a inversao da polaridade dos bits do RX
	uint8_t s2_rwuid;		// habilitar o set do IDLE bit durante standby do RX
	uint8_t s2_brk13;		// selecionar o comprimento de caracter break (0=10 bits; 1=13 bits)
	uint8_t s2_lbkde;		// habilitar deteccao de caractere break longo
	uint8_t c3_r8t9; 		// bit 8 (RX)/bit 9 (TX)
	uint8_t c3_r9t8;		// bit 9 (RX)/bit 8 (TX)
	uint8_t c3_txdir; 		// na configuracao RSRC=1, configurar o sentido de TX (0=entrada; 1=saida)
	uint8_t c3_txinv;		// habilitar a inversao da polaridade dos bits de TX
	uint8_t c4_maen1;		// habilitar controle de "match address" 1
	uint8_t c4_maen2; 		// habilitar controle de "match address" 2
	uint8_t c4_m10;			// selecionar o modo de bits (0=8/9 bits;1=10 bits)
	uint8_t c4_osr;			// taxa de super-amostragem (default=16(0b01111)
	uint8_t c5_tdmae; 		// habilitar transmissao por DMA
	uint8_t c5_rdmae; 		// habilitar recepcao por DMA
	uint8_t c5_bothedge; 	// amostrar os dados em ambas as bordas do clock de baud rate
	uint8_t c5_resyncdis;	// desabilitar o resincronismo nas transies de 1 para 0
} UART0Config;

/**
 * @brief Inicializar os pinos conectados a UART0 e ao OpenSDA.
 */
void UART0_init(UART0Config *config) {
	SIM_SCGC4 |= SIM_SCGC4_UART0_MASK;      // habilitar clock gate de UART0
	
	SIM_SOPT2 &= ~SIM_SOPT2_UART0SRC(0b11); // configurar a fonte de relgio (20.971520MHz)
	SIM_SOPT2 |= SIM_SOPT2_UART0SRC(0x1);	
	
	// Desabilitar canais RX e TX para configuracoes
	UART0_C2 &= ~(UART0_C2_RE_MASK |           
				UART0_C2_TE_MASK);
	
	// Configurar stop bits
	UART0_BDH &= ~(UART0_BDH_SBNS_MASK);
	UART0_BDH |= config->bdh_sbns << UART0_BDH_SBNS_SHIFT;
	
	// Configurar o registrador C1
	UART0_C1 &= ~(UART0_C1_LOOPS_MASK |
			UART0_C1_DOZEEN_MASK |
			UART0_C1_RSRC_MASK |
			UART0_C1_M_MASK |                 
			UART0_C1_WAKE_MASK |
			UART0_C1_ILT_MASK |
			UART0_C1_PE_MASK |			
			UART0_C1_PT_MASK
			 );

	UART0_C1 |= (config->c1_loops << UART0_C1_LOOPS_SHIFT |
			config->c1_dozeen << UART0_C1_DOZEEN_SHIFT |
			config->c1_rsrc << UART0_C1_RSRC_SHIFT |
			config->c1_m << UART0_C1_M_SHIFT |                 
			config->c1_wake << UART0_C1_WAKE_SHIFT |
			config->c1_ilt << UART0_C1_ILT_SHIFT |
			config->c1_pe << UART0_C1_PE_SHIFT |			
			config->c1_pt << UART0_C1_PT_SHIFT
			 );
	
	// Configurar o registrador C2
	UART0_C2 &= ~(UART0_C2_RWU_MASK |
			UART0_C2_SBK_MASK);
	
	UART0_C2 |= (config->c2_rwu << UART0_C2_RWU_SHIFT |
			config->c2_sbk << UART0_C2_SBK_SHIFT
			);
	
	// Configurar o registrador S2
	UART0_S2 &= ~(UART0_S2_MSBF_MASK |
			UART0_S2_RXINV_MASK |
			UART0_S2_RWUID_MASK |
			UART0_S2_BRK13_MASK |
			UART0_S2_LBKDE_MASK
			);
	
	UART0_S2 |= (config->s2_msbf << UART0_S2_MSBF_SHIFT |
			config->s2_rxinv << UART0_S2_RXINV_SHIFT |
			config->s2_rwuid << UART0_S2_RWUID_SHIFT |
			config->s2_brk13 << UART0_S2_BRK13_SHIFT |
			config->s2_lbkde << UART0_S2_LBKDE_SHIFT
			);
	
	// Configurar o registrador C3
	UART0_C3 &= ~(UART0_C3_R8T9_MASK |
			UART0_C3_R9T8_MASK |
			UART0_C3_TXDIR_MASK |
			UART0_C3_TXINV_MASK 
			);
	
	UART0_C3 |= (config->c3_r8t9 << UART0_C3_R8T9_SHIFT |
			config->c3_r9t8 << UART0_C3_R9T8_SHIFT |
			config->c3_txdir << UART0_C3_TXDIR_SHIFT |
			config->c3_txinv << UART0_C3_TXINV_SHIFT 
			);

	// Configurar o registrador C4
	UART0_C4 &= ~(UART0_C4_MAEN1_MASK |
			UART0_C4_MAEN2_MASK |
			UART0_C4_M10_MASK
			);
	
	UART0_C4 |= (config->c4_maen1 << UART0_C4_MAEN1_SHIFT |
			config->c4_maen2 << UART0_C4_MAEN2_SHIFT |
			config->c4_m10 << UART0_C4_M10_SHIFT
			);
	
	// Valores fora do intervalo 4 a 32 implicam em setar 16
	UART0_C4 |= (0b11111 << UART0_C4_OSR_SHIFT);
	UART0_C4 &= ((config->c4_osr << UART0_C4_OSR_SHIFT) | ~UART0_C4_OSR_MASK);

				
	// Configurar o registrador C5				
	UART0_C5 &= ~(UART0_C5_TDMAE_MASK |
			UART0_C5_RDMAE_MASK |
			UART0_C5_BOTHEDGE_MASK |
			UART0_C5_RESYNCDIS_MASK
			);
	
	UART0_C5 |= (config->c5_tdmae << UART0_C5_TDMAE_SHIFT |
			config->c5_rdmae << UART0_C5_RDMAE_SHIFT |
			config->c5_bothedge << UART0_C5_BOTHEDGE_SHIFT |
			config->c5_resyncdis << UART0_C5_RESYNCDIS_SHIFT
			);

	// Ajustar amostragem conforme a taxa de superamostragem
	if ((config->c4_osr & 0b11111) > 2 && (config->c4_osr & 0b11111) < 7)
		UART0_C5 |= UART0_C5_BOTHEDGE_MASK;    
	
	// Mascarar bits invalidos
	config->sbr |= (config->sbr & 0x1FFF);
	
	// Setar baus rate via SBR
	UART0_BDH &= ~UART0_BDH_SBR(0x1F);                             
	UART0_BDL &= ~UART0_BDL_SBR(0xFF);            
	UART0_BDH |= UART0_BDH_SBR(config->sbr >> 8);                             
	UART0_BDL |= UART0_BDL_SBR(config->sbr & 0x00FF);            
		
	//Habilitar os canais RX e TX
	UART0_C2 |= (UART0_C2_RE_MASK |         		///< habilita o canal receptor 
				 UART0_C2_TE_MASK);	  			///< habilita o canal transmissor  
	
	return;
}

/**
 * @brief Reconfigurar UART0
 */
void UART0_reconfig(UART0Config *config) {
	// Desabilitar canais RX e TX para configuracoes
	UART0_C2 &= ~(UART0_C2_RE_MASK |           
				UART0_C2_TE_MASK);
	
	// Configurar stop bits
	UART0_BDH &= ~(UART0_BDH_SBNS_MASK);
	UART0_BDH |= config->bdh_sbns << UART0_BDH_SBNS_SHIFT;
	
	// Configurar o registrador C1
	UART0_C1 &= ~(UART0_C1_LOOPS_MASK |
			UART0_C1_DOZEEN_MASK |
			UART0_C1_RSRC_MASK |
			UART0_C1_M_MASK |                 
			UART0_C1_WAKE_MASK |
			UART0_C1_ILT_MASK |
			UART0_C1_PE_MASK |			
			UART0_C1_PT_MASK
			 );

	UART0_C1 |= (config->c1_loops << UART0_C1_LOOPS_SHIFT |
			config->c1_dozeen << UART0_C1_DOZEEN_SHIFT |
			config->c1_rsrc << UART0_C1_RSRC_SHIFT |
			config->c1_m << UART0_C1_M_SHIFT |                 
			config->c1_wake << UART0_C1_WAKE_SHIFT |
			config->c1_ilt << UART0_C1_ILT_SHIFT |
			config->c1_pe << UART0_C1_PE_SHIFT |			
			config->c1_pt << UART0_C1_PT_SHIFT
			 );
	
	// Configurar o registrador C2
	UART0_C2 &= ~(UART0_C2_RWU_MASK |
			UART0_C2_SBK_MASK);
	
	UART0_C2 |= (config->c2_rwu << UART0_C2_RWU_SHIFT |
			config->c2_sbk << UART0_C2_SBK_SHIFT
			);
	
	// Configurar o registrador S2
	UART0_S2 &= ~(UART0_S2_MSBF_MASK |
			UART0_S2_RXINV_MASK |
			UART0_S2_RWUID_MASK |
			UART0_S2_BRK13_MASK |
			UART0_S2_LBKDE_MASK
			);
	
	UART0_S2 |= (config->s2_msbf << UART0_S2_MSBF_SHIFT |
			config->s2_rxinv << UART0_S2_RXINV_SHIFT |
			config->s2_rwuid << UART0_S2_RWUID_SHIFT |
			config->s2_brk13 << UART0_S2_BRK13_SHIFT |
			config->s2_lbkde << UART0_S2_LBKDE_SHIFT
			);
	
	// Configurar o registrador C3
	UART0_C3 &= ~(UART0_C3_R8T9_MASK |
			UART0_C3_R9T8_MASK |
			UART0_C3_TXDIR_MASK |
			UART0_C3_TXINV_MASK 
			);
	
	UART0_C3 |= (config->c3_r8t9 << UART0_C3_R8T9_SHIFT |
			config->c3_r9t8 << UART0_C3_R9T8_SHIFT |
			config->c3_txdir << UART0_C3_TXDIR_SHIFT |
			config->c3_txinv << UART0_C3_TXINV_SHIFT 
			);

	// Configurar o registrador C4
	UART0_C4 &= ~(UART0_C4_MAEN1_MASK |
			UART0_C4_MAEN2_MASK |
			UART0_C4_M10_MASK
			);
	
	UART0_C4 |= (config->c4_maen1 << UART0_C4_MAEN1_SHIFT |
			config->c4_maen2 << UART0_C4_MAEN2_SHIFT |
			config->c4_m10 << UART0_C4_M10_SHIFT
			);
	
	// Valores fora do intervalo 4 a 32 implicam em setar 16
	UART0_C4 |= (0b11111 << UART0_C4_OSR_SHIFT);
	UART0_C4 &= ((config->c4_osr << UART0_C4_OSR_SHIFT) | ~UART0_C4_OSR_MASK);
				
	// Configurar o registrador C5				
	UART0_C5 &= ~(UART0_C5_TDMAE_MASK |
			UART0_C5_RDMAE_MASK |
			UART0_C5_BOTHEDGE_MASK |
			UART0_C5_RESYNCDIS_MASK
			);
	
	UART0_C5 |= (config->c5_tdmae << UART0_C5_TDMAE_SHIFT |
			config->c5_rdmae << UART0_C5_RDMAE_SHIFT |
			config->c5_bothedge << UART0_C5_BOTHEDGE_SHIFT |
			config->c5_resyncdis << UART0_C5_RESYNCDIS_SHIFT
			);

	// Ajustar amostragem conforme a taxa de superamostragem
	if ((config->c4_osr & 0b11111) > 2 && (config->c4_osr & 0b11111) < 7)
		UART0_C5 |= UART0_C5_BOTHEDGE_MASK;    
	
	// Mascarar bits invalidos
	config->sbr |= (config->sbr & 0x1FFF);
	
	// Setar baus rate via SBR
	UART0_BDH &= ~UART0_BDH_SBR(0x1F);                             
	UART0_BDL &= ~UART0_BDL_SBR(0xFF);            
	UART0_BDH |= UART0_BDH_SBR(config->sbr >> 8);                             
	UART0_BDL |= UART0_BDL_SBR(config->sbr & 0x00FF);            
		
	//Habilitar os canais RX e TX
	UART0_C2 |= (UART0_C2_RE_MASK |         		///< habilita o canal receptor 
				 UART0_C2_TE_MASK);	  			///< habilita o canal transmissor  
	
	return;
}

/**
 * @brief Habilitar o canal receptor
 */
void UART0_enableRE (void) {
	UART0_C2 |= UART_C2_RE_MASK;
	
}

/**
 * @brief Habilitar o canal transmissor
 */
void UART0_enableTE (void) {
	UART0_C2 |= UART_C2_TE_MASK;
}

/**
 * @brief Desabilitar o canal receptor (RX)
 */
void UART0_disableRE (void) {
	UART0_C2 &= ~UART_C2_RE_MASK;
}

/**
 * @brief Desabilitar o canal transmissor (TX)
 */
void UART0_disableTE (void) {
	UART0_C2 &= ~UART_C2_TE_MASK;
	
}

//=====================================
// Application
//=====================================
UART0Config config = {
		.bdh_sbns=0,			// Um stop bit
		.sbr=0b100, 			// MCGFLLCLK/(baud_rate*osr)=4
		.c1_loops=0,			// Operacao normal: full-duplex 
		.c1_dozeen=0,			// UART habilitado no modo WAIT
		.c1_rsrc=0,				// nao tem efeito para loops == 0
		.c1_m=0,				// dados em 8 bits
		.c1_wake=0,				// wakeup de RX por linha ociosa (IDLE)
		.c1_ilt=0,				// deteccao do estado ocioso pela contagem apos start bits
		.c1_pe=0,				// nao ha checagem de paridade por hardware
		.c1_pt=0,				// nao tem efeito para pe==0
		.c2_rwu=0,				// operacao normal de recepcao do UART0
		.c2_sbk=0,				// operacao normal de transmissor do UART0
		.s2_msbf=0,				// envio de LSB para MSB
		.s2_rxinv=0,			// polaridade dos dados de RX nao eh invertida
		.s2_rwuid=0,			// nao tem efeito para rwu==0
		.s2_brk13=0,			// nao tem efeito para operacao normal de transmissor
		.s2_lbkde=0,			// deteccao do caractere break no tamanho menor
		.c3_r8t9=0,				// nao tem efeito para m==0
		.c3_r9t8=0,				// nao tem efeito para m==0
		.c3_txdir=0,			// nao tem efeito para loops==0
		.c3_txinv=0,			// polaridade dos dados de TX nao eh invertida 
		.c4_maen1=0,			// todos os dados recebidos sao transferidos para buffer de dados
		.c4_maen2=0,			
		.c4_m10=0,				// tamanho dos dados em RX e TX  8 ou 9 bits
		.c4_osr=0b01111,		// superamostragem 16x
		.c5_tdmae=0,			// uso de DMA pelo TX desabilitada
		.c5_rdmae=0,			// uso de DMA pelo RX desabilitado
		.c5_bothedge=0,			// amostragem somente na borda de subida do clock
		.c5_resyncdis=0			// resincronizacao na recepcao habilitada
};

// Baud rates: 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200
#define BAUD_RATE 115200

int main (void)
{
	SIM_initBusClock (0b000);	// Divisor de frequencia em 1
	SIM_selPLLFLL (0); 			// selecionar MCGFLLCLK (20971520Hz)
	PORTx_muxUART0RxTxPins();
	
	config.c4_osr = 15;
	config.sbr = (uint16_t)(20971520./(BAUD_RATE * (config.c4_osr+1)));
	if ((20971520./(1.0 * BAUD_RATE * (config.c4_osr+1))) > config.sbr)
		config.sbr++;
	config.c1_pe=1;				// nao ha checagem de paridade por hardware
	config.c1_pt=0;				// paridade par
	config.c1_m=1;				// bit de paridade eh 9o. bit

	UART0_init(&config);
	
	char d;
	
	for(;;) {
		if (UART0_S1 & UART0_S1_RDRF_MASK) {
			//Ler o dado no receptor
			d = UART0_D;   
			while (!(UART0_S1 & UART0_S1_TDRE_MASK));
			//Ecoar o dado no transmissor
			UART0_D = d;				
		}
	}	
	
	return 0;
}
